/* $Header: /src/carson/nistnet/monitor/RCS/hitmonitor.c,v 1.4 1998/01/30 17:05:00 carson Exp $ */

/* hitmonitor.c -- NIST net network emulator X-based user interface.
 * This version is written to use the Athena widget set, with provisions
 * for use with neXtaw as well.  I've also successfully linked it with
 * the Xaw3d widget set.  To my eyes, neXtaw gives the most comprehensible
 * user interface (the scroll bars in particular), but tastes may vary.
 *
 * To use, the nistnet kernel module must be installed.  The user must also
 * be root (in order to open the nistnet device).
 *
 * This command takes the usual X toolkit arguments.
 */
#include "xincludes.h"

/* These are for neXtaw */
#ifndef XtNdrawBorder
#define XtNdrawBorder "drawBorder"
#endif
#ifndef XtNdrawArrows
#define XtNdrawArrows "drawArrows"
#endif
#ifndef XtNshadowWidth
#define XtNshadowWidth "shadowWidth"
#endif

#include "nistnet.xpm"

/*****************************Definitions***********************************/

#define MAX_ROWS	50
#define PER_SECOND	4

/* Resource names for resource database */
#define XtNdebug        "debug"         /*  resource name for debug flag */
#define XtCDebug        "Debug"         /*  resource class for debug flag */


/*****************************Global Data***********************************/
Display *theDisplay;                    /*  the display */


typedef struct {                        /*  application context data */
	Boolean Debug;                  /*  debug mode (unused) */
} AppData;

AppData app_data;

static XtResource ResourceList[] = {
        {                               /*  debugging flag */
        XtNdebug,               /*  name = "debug" */
        XtCDebug,               /*  class = "Debug" */
        XtRBoolean,             /*  representation type = "Boolean" */
        sizeof(Boolean),        /*  size of entry in AppData */
        XtOffsetOf(AppData, Debug),     /*  offset in AppData */
        XtRImmediate,           /*  next argument is value */
        (XtPointer) FALSE,      /*  default = FALSE (0) */
        },

};

static XrmOptionDescRec Options[] = {
        {"-debug",         "*debug",              XrmoptionNoArg, "True"},
};



#define	THOUSAND	1000	/* so as not to mess up typing */

typedef enum {
	NullValue=0,
	IntegerValue,
	UsecValue,
	ShortpercentValue,
	StringValue
} Value;

typedef struct {
	char *name;
	Value type;
} TableValue;

TableValue row1[] = {
	{"Source", StringValue},
	{"Dest", StringValue},
	{"Delay (ms)", UsecValue},
	{"Delsigma(ms)", UsecValue},
	{"Bandwidth", IntegerValue},
	{"Drop %", ShortpercentValue},
	{"Dup %", ShortpercentValue},
	{"DRDmin", IntegerValue},
	{"DRDmax", IntegerValue},
	{NULL, NullValue}
};

struct value_stats {
	int changed;
	int inuse;
	struct lin_hitstats stats;
};


struct value_widget {
	struct value_stats *changeme;
	Widget widget;
	Value type;
	union {
		int number;
		char *string;
	} value;
};

int table_size=10;

struct value_stats  hitstats[MAX_ROWS];

struct value_widget row1_values[9][MAX_ROWS];

TableValue row2[] = {
	{"AvBandwidth", IntegerValue},
	{"Drops", IntegerValue},
	{"Dups", IntegerValue},
	{"PacketTime", UsecValue},
	{"PacketSize", IntegerValue},
	{"QueueLength", IntegerValue},
	{"BytesSent", IntegerValue},
	{NULL, NullValue}
};

struct value_widget row2_values[7][MAX_ROWS];

/***************************Function Declaration****************************/
/*  Functions invoked by callbacks and actions */
static void On(Widget w, XtPointer client_data, XtPointer call_data);
static void Off(Widget w, XtPointer client_data, XtPointer call_data);
static void Update(Widget w, XtPointer client_data, XtPointer call_data);
static void Quit(Widget w, XtPointer client_data, XtPointer call_data);
static void QuitAction(Widget w, XEvent *event, String *params, Cardinal *num);

static void get_text(Widget w, XtPointer client_data, XtPointer call_data);
static void MyXawTextSetString(Widget textwidget, char *string);
static void MyXawTextSetNumber(Widget textwidget, unsigned int value, Value type, Boolean justify);

/*void UpdateTextPosition();*/

static XtActionsRec Window_actions[] = {
        {"Quit",                        QuitAction},
};

Widget
MakeLabelWidget(char *text, Arg *position, struct value_widget *entry,
	Dimension textwidth, Boolean can_edit, Widget form)
{
	Arg textargs[20];
	int n=0;
	Widget answer;

	textargs[0] = position[0];
	textargs[1] = position[1];
	textargs[2] = position[2];
	textargs[3] = position[3]; n=4;
	if (entry) {
		XtSetArg(textargs[n], XtNstring, entry->value.string); n++;
	} else {
		XtSetArg(textargs[n], XtNstring, text); n++;
	}
	XtSetArg(textargs[n], XtNwidth, textwidth); n++;
	XtSetArg(textargs[n], XtNborderWidth, 0); n++;
	if (!can_edit) {	/* i.e. uneditable text field */
		/* anything special? */
	} else {
		/* to compensate for width */
		/* should be able to do from some sort of defines,
		 * defaultBorderWidth + defaultInternalWidth, grumble
		 */
		XtSetArg(textargs[n], XtNinternalWidth, 6); n++;
	}
	/* Keep widgets from being resized and generally messing
	 * up the form....
	 */
	XtSetArg(textargs[n], XtNleft, XawChainLeft); n++;
	XtSetArg(textargs[n], XtNright, XawChainLeft); n++;
	XtSetArg(textargs[n], XtNtop, XawChainTop); n++;
	XtSetArg(textargs[n], XtNbottom, XawChainTop); n++;
	answer =  XtCreateManagedWidget(text, labelWidgetClass, form,
		textargs, n);
	if (!answer) {
		perror(text);
		return NULL;
	}
	return answer;
}

Widget
MakeTextWidget(char *text, Arg *position, struct value_widget *entry,
	Dimension textwidth, Boolean can_edit, Widget form)
{
	Arg textargs[20];
	int n=0;
	Widget answer;

	textargs[0] = position[0];
	textargs[1] = position[1];
	textargs[2] = position[2];
	textargs[3] = position[3]; n=4;
	if (can_edit) {
		XtSetArg(textargs[n], XtNeditType, XawtextEdit); n++;
	} else {
		XtSetArg(textargs[n], XtNdisplayCaret, FALSE); n++;
		XtSetArg(textargs[n], XtNborderWidth, 0); n++;
	}
	XtSetArg(textargs[n], XtNwidth, textwidth); n++;
	/* Keep widgets from being resized and generally messing
	 * up the form....
	 */
	XtSetArg(textargs[n], XtNleft, XawChainLeft); n++;
	XtSetArg(textargs[n], XtNright, XawChainLeft); n++;
	XtSetArg(textargs[n], XtNtop, XawChainTop); n++;
	XtSetArg(textargs[n], XtNbottom, XawChainTop); n++;
	answer =  XtCreateManagedWidget(text, asciiTextWidgetClass, form,
		textargs, n);
	if (!answer) {
		perror(text);
		return NULL;
	}
	if (can_edit)
		XtAddCallback(XawTextGetSource(answer), XtNcallback,
			get_text, (XtPointer) entry);
	switch (entry->type) {
	case IntegerValue:
	case UsecValue:
	case ShortpercentValue:
		MyXawTextSetNumber(answer, entry->value.number, entry->type, TRUE);
		break;
	case StringValue:
		MyXawTextSetString(answer, entry->value.string);
	}
	return answer;
}

static Arg position_args[] = {
	{XtNfromVert, (XtArgVal) NULL},
	{XtNfromHoriz, (XtArgVal) NULL},
	{XtNhorizDistance, (XtArgVal) 5},
	{XtNvertDistance, (XtArgVal) 5}
};

static void
Remove(Widget w, XtPointer client_data, XtPointer call_data)
{
	int i, table = (int)client_data;

	hitstats[table].changed = FALSE;
	hitstats[table].inuse = FALSE;
	if (rmhit(&hitstats[table].stats.hitreq) < 0) {
		fprintf(stderr, "Error removing %s -> %s: ",
			row1_values[0][table].value.string,
			row1_values[1][table].value.string);
		perror("rmhit");
	}
	bzero(&hitstats[table].stats, sizeof(struct lin_hitstats));
	free(row1_values[0][table].value.string);
	row1_values[0][table].value.string = (char *)
		malloc(50);
	row1_values[0][table].value.string[0] = '\0';
	free(row1_values[1][table].value.string);
	row1_values[1][table].value.string = (char *)
		malloc(50);
	row1_values[1][table].value.string[0] = '\0';
	MyXawTextSetString(row1_values[0][table].widget,
		row1_values[0][table].value.string);
	MyXawTextSetString(row1_values[1][table].widget,
		row1_values[1][table].value.string);
	for (i= 2; row1[i].name; ++i) {
		row1_values[i][table].value.number = 0;
		MyXawTextSetNumber(row1_values[i][table].widget, 0, row1[i].type, TRUE);
	}
	for (i=0; row2[i].name; ++i) {
		row2_values[i][table].value.number = 0;
		MyXawTextSetNumber(row2_values[i][table].widget, 0, row2[i].type, TRUE);
	}

}

Widget
FillFormHead(Widget formwidget, Boolean lefthand)
{
	Widget vertguide = NULL;
	Widget horizguide = NULL, newwidget;
	int i, table;

#define FORM_torightof(wid, toleft)	wid[1].value = (XtArgVal)toleft
#define FORM_below(wid, above)		wid[0].value = (XtArgVal)above


	horizguide = NULL;
	table = 0;
	for (i= (lefthand ? 0 : 2); lefthand ? (i < 2) : (int)row1[i].name; ++i) {
		FORM_torightof(position_args, horizguide);
		FORM_below(position_args, vertguide);
		row1_values[i][table].value.string = (char *)
				malloc(strlen(row1[i].name)+1);
		strcpy(row1_values[i][table].value.string,
				row1[i].name);
		row1_values[i][table].changeme = NULL;
		row1_values[i][table].type = StringValue;
		row1_values[i][table].widget = newwidget =
			MakeLabelWidget(row1[i].name, position_args,
				&row1_values[i][table], 
				lefthand ? 175 : 133,
				TRUE, formwidget);
		horizguide = newwidget;
	}
	if (lefthand) {		/* pad out to account for "remove" button */
		Widget Wremove;
		Arg removeargs[20];
		int n;

		FORM_torightof(position_args, horizguide);
		FORM_below(position_args, vertguide);
		removeargs[0] = position_args[0];
		removeargs[1] = position_args[1];
		removeargs[2] = position_args[2];
		removeargs[3] = position_args[3]; n=4;
		/* Keep widgets from being resized and generally messing
		 * up the form....
		 */
		XtSetArg(removeargs[n], XtNleft, XawChainLeft); n++;
		XtSetArg(removeargs[n], XtNright, XawChainLeft); n++;
		XtSetArg(removeargs[n], XtNtop, XawChainTop); n++;
		XtSetArg(removeargs[n], XtNbottom, XawChainTop); n++;
		XtSetArg(removeargs[n], XtNborderWidth, 0); n++;
		Wremove = XtCreateManagedWidget("Remove", labelWidgetClass, formwidget, removeargs, n);
		XtVaSetValues(Wremove, XtNlabel, "      ", NULL);
		return;
	}
	for (i=0; row2[i].name; ++i) {
		FORM_torightof(position_args, horizguide);
		FORM_below(position_args, vertguide);
		row2_values[i][table].value.string = (char *)
				malloc(strlen(row2[i].name)+1);
		strcpy(row2_values[i][table].value.string,
				row2[i].name);
		row2_values[i][table].changeme = NULL;
		row2_values[i][table].type = StringValue;
		row2_values[i][table].widget = newwidget =
			MakeLabelWidget(row2[i].name, position_args,
				&row2_values[i][table],
				133, FALSE, formwidget);
		horizguide = newwidget;
	}
	return newwidget;
}


static Widget vertguide =  NULL;

void
AddNewRow(Widget formwidget, int table, Boolean lefthand)
{
	int i;
	Widget newwidget;
	Widget horizguide;

	horizguide = NULL;
	/* Number the rows...
	{
	char number[20];

	FORM_torightof(position_args, horizguide);
	FORM_below(position_args, vertguide);
	sprintf(number, "%2d", table);
	horizguide = MakeLabelWidget(number, position_args, NULL,
		25, FALSE, formwidget);
	}
	*/

	for (i = (lefthand ? 0 : 2); lefthand ? (i < 2) : (int)row1[i].name; ++i) {
		FORM_torightof(position_args, horizguide);
		FORM_below(position_args, vertguide);
		if (lefthand) {
			row1_values[i][table].value.string = (char *)
				malloc(50);
			row1_values[i][table].value.string[0] = '\0';
		}
		row1_values[i][table].changeme = &hitstats[table];
		row1_values[i][table].type = row1[i].type;
		row1_values[i][table].widget = newwidget =
			MakeTextWidget(row1[i].name, position_args,
				&row1_values[i][table],
				lefthand ? 175 : 133,
				TRUE, formwidget);
		horizguide = newwidget;
	}
	if (lefthand) {		/* add a "remove" button */
		Widget Wremove;
		Arg removeargs[20];
		int n;

		FORM_torightof(position_args, horizguide);
		FORM_below(position_args, vertguide);
		removeargs[0] = position_args[0];
		removeargs[1] = position_args[1];
		removeargs[2] = position_args[2];
		removeargs[3] = position_args[3]; n=4;
		/* Keep widgets from being resized and generally messing
		 * up the form....
		 */
		XtSetArg(removeargs[n], XtNleft, XawChainLeft); n++;
		XtSetArg(removeargs[n], XtNright, XawChainLeft); n++;
		XtSetArg(removeargs[n], XtNtop, XawChainTop); n++;
		XtSetArg(removeargs[n], XtNbottom, XawChainTop); n++;
		Wremove = XtCreateManagedWidget("Remove", commandWidgetClass, formwidget, removeargs, n);
		XtAddCallback(Wremove, XtNcallback, Remove, (XtPointer)table);
	} else {
		for (i=0; row2[i].name; ++i) {
			FORM_torightof(position_args, horizguide);
			FORM_below(position_args, vertguide);
			row2_values[i][table].changeme = &hitstats[table];
			row2_values[i][table].type = row2[i].type;
			row2_values[i][table].widget = newwidget =
				MakeTextWidget(row2[i].name, position_args,
					&row2_values[i][table],
					133, FALSE, formwidget);
			horizguide = newwidget;
		}
	}
	vertguide = newwidget;
}


XtAppContext app_context;

#define UpdateValue1(col, i, source, always) \
	if (always || row1_values[col][i].value.number != hitstats[i].stats.hitreq.source) {\
		MyXawTextSetNumber(row1_values[col][i].widget, hitstats[i].stats.hitreq.source, row1[col].type, TRUE);\
		row1_values[col][i].value.number = hitstats[i].stats.hitreq.source; \
	}
#define UpdateValue2(col, i, source, always) \
	if (always || row2_values[col][i].value.number != hitstats[i].stats.source) {\
		MyXawTextSetNumber(row2_values[col][i].widget, hitstats[i].stats.source, row2[col].type, TRUE);\
		row2_values[col][i].value.number = hitstats[i].stats.source; \
	}

RedrawAll(Boolean always)
{
	int i;
	unsigned time_no_see;

	for (i=1; i < table_size; ++i) {
		/* Don't touch settable values if being edited */
		if (hitstats[i].inuse && !hitstats[i].changed) {
			if (stathit(&hitstats[i].stats) < 0) {
				fprintf(stderr, "Can't stat %s -> %s: ",
					row1_values[0][i].value.string,
					row1_values[1][i].value.string);
				perror("stathit");
			}
			UpdateValue1(2, i, delay, always);
			UpdateValue1(3, i, delsigma, always);
			UpdateValue1(4, i, bandwidth, always);
			UpdateValue1(5, i, p_drop, always);
			UpdateValue1(6, i, p_dup, always);
			UpdateValue1(7, i, drd_min, always);
			UpdateValue1(8, i, drd_max, always);

			UpdateValue2(0, i, current_bandwidth, always);
			UpdateValue2(1, i, n_drops, always);
			UpdateValue2(2, i, dups, always);
			/* Time, ugh... */
			time_no_see = hitstats[i].stats.last_packet.tv_sec*MILLION+
					hitstats[i].stats.last_packet.tv_usec;
			if (row2_values[3][i].value.number != time_no_see) {
				MyXawTextSetNumber(row2_values[3][i].widget, time_no_see, UsecValue, TRUE);
				row2_values[3][i].value.number = time_no_see;
			}
			UpdateValue2(4, i, current_size, always);
			UpdateValue2(5, i, qlen, always);
			UpdateValue2(6, i, bytes_sent, always);
		}
	}
}

static void
TimeoutHandler(Widget w, XtIntervalId *id)
{

	/* Reprime ourselves */
	XtAppAddTimeOut(app_context, 1000/PER_SECOND,
		(XtTimerCallbackProc)TimeoutHandler, NULL);

	RedrawAll(FALSE);

}

static Widget formleft, formright;

static void AddNumberedRow(int number)
{
	Widget oldguide = vertguide;

	AddNewRow(formleft, number, TRUE);
	vertguide = oldguide;
	AddNewRow(formright, number, FALSE);
}

void
FillForm(Widget titleformleft, Widget titleformright, int table_size)
{
	Widget newwidget;
	int table, i;

	(void) FillFormHead(titleformleft, TRUE);
	(void) FillFormHead(titleformright, FALSE);
	for (table=1; table < table_size; ++table) {
		AddNumberedRow(table);
	}

}

static void
AddRow(Widget w, XtPointer client_data, XtPointer call_data)
{
	if (table_size < MAX_ROWS) {
		AddNumberedRow(table_size);
		++table_size;
	}
}


/* ReadHit -- read in all current hit table entries.  Force a display
 * update too (not strictly required).
 */
static void
ReadHit(Widget w, XtPointer client_data, XtPointer call_data)
{
	int i, count, ret;
	struct addrpair bigguy[MAX_ROWS];

	count = readhit(bigguy, MAX_ROWS-1);
	while (table_size < count+1)
		AddRow(w, client_data, call_data);
	for (i=1; i < count+1; ++i) {
		hitstats[i].changed = TRUE;
		hitstats[i].stats.hitreq.src = bigguy[i-1].src;
		hitstats[i].stats.hitreq.dest = bigguy[i-1].dest;
		hitstats[i].inuse = TRUE;
		fixname(row1_values[0][i].value.string,
			hitstats[i].stats.hitreq.src);
		MyXawTextSetString(row1_values[0][i].widget,
			row1_values[0][i].value.string);
		fixname(row1_values[1][i].value.string,
			hitstats[i].stats.hitreq.dest);
		MyXawTextSetString(row1_values[1][i].widget,
			row1_values[1][i].value.string);
		hitstats[i].changed = FALSE;
	}
	RedrawAll(TRUE);
}

main(int argc, char **argv)
{
	Widget topLevel;
	Widget title;
	Pixmap titlepix;
	Widget bigpane, mainpane, leftpane, rightpane;
	Widget titleviewleft, titleviewright;
	Widget titleformleft, titleformright;
	Widget viewleft, viewright;
	Widget buttonbox;
	Widget Wupdate;
	Widget Won;
	Widget Woff;
	Widget Wread;
	Widget Wadd;
	Widget Wquit;
	int theScreen;
	Arg args[20];
	int n;
	unsigned int nistnet_width, nistnet_height;

	if (openhit() < 0) {
		perror("Hitbox device (/dev/hitbox):");
		exit(1);
	}
	argv[0] = "NIST Network Emulation Tool v. 1.0";
	topLevel = XtVaAppInitialize(&app_context, "nistnet",
		Options, XtNumber(Options),
		&argc, argv,
		NULL, NULL);
	theDisplay = XtDisplay(topLevel);
	theScreen = DefaultScreen(theDisplay);
	titlepix = XMineCreatePixmapFromData(theDisplay,
		RootWindow(theDisplay, theScreen),
		nistnet, &nistnet_width, &nistnet_height);
	if (!titlepix)
		perror("create bitmap");
	XtGetApplicationResources(
		topLevel, &app_data, ResourceList, XtNumber(ResourceList),
		NULL, 0);
	n = 0;
	bigpane = XtCreateManagedWidget("pane", panedWidgetClass, topLevel, args, n);
	n = 0;
	XtSetArg(args[n], XtNheight, nistnet_height); n++;
	XtSetArg(args[n], XtNwidth, nistnet_width); n++;
	XtSetArg(args[n], XtNbitmap, titlepix); n++;
	XtSetArg(args[n], XtNbackground, BlackPixel(theDisplay, theScreen)); n++;
	title = XtCreateManagedWidget("title", labelWidgetClass, bigpane, args, n);
	n = 0;
	XtSetArg(args[n], XtNorientation, XtorientHorizontal); n++;
	mainpane = XtCreateManagedWidget("mainpane", panedWidgetClass, bigpane, args, n);
	n = 0;
	leftpane = XtCreateManagedWidget("leftpane", panedWidgetClass, mainpane, args, n);
	rightpane = XtCreateManagedWidget("rightpane", panedWidgetClass, mainpane, args, n);
	n = 0;
	XtSetArg(args[n], XtNallowHoriz, TRUE); n++;
	XtSetArg(args[n], XtNforceBars, TRUE); n++;
	XtSetArg(args[n], XtNallowVert, TRUE); n++;
	/* Really only needed for neXtaw... */
	XtSetArg(args[n], XtNheight, 44); n++;
	titleviewleft = XtCreateManagedWidget("titleviewleft", viewportWidgetClass, leftpane, args, n);
	--n;
	XtSetArg(args[n], XtNuseBottom, TRUE); n++;
	viewleft = XtCreateManagedWidget("viewleft", viewportWidgetClass, leftpane, args, n);


	n = 0;
	titleformleft = XtCreateManagedWidget("titleformleft", formWidgetClass, titleviewleft, args, n);
	formleft = XtCreateManagedWidget("formleft", formWidgetClass, viewleft, args, n);
	n = 0;
	XtSetArg(args[n], XtNallowHoriz, TRUE); n++;
	XtSetArg(args[n], XtNforceBars, TRUE); n++;
	XtSetArg(args[n], XtNwidth, 700); n++;
	XtSetArg(args[n], XtNallowVert, TRUE); n++;
	/* Really only needed for neXtaw... */
	XtSetArg(args[n], XtNheight, 44); n++;
	titleviewright = XtCreateManagedWidget("titleviewright", viewportWidgetClass, rightpane, args, n);
	--n;
	XtSetArg(args[n], XtNuseBottom, TRUE); n++;
	viewright = XtCreateManagedWidget("viewright", viewportWidgetClass, rightpane, args, n);
	n = 0;
	titleformright = XtCreateManagedWidget("titleformright", formWidgetClass, titleviewright, args, n);
	formright = XtCreateManagedWidget("formright", formWidgetClass, viewright, args, n);

	FillForm(titleformleft, titleformright, table_size);

	n = 0;
	XtSetArg(args[n], XtNskipAdjust, TRUE); n++;
	XtSetArg(args[n], XtNhSpace, 10); n++;
	buttonbox = XtCreateManagedWidget("buttonbox", boxWidgetClass, bigpane, args, n);
	n = 0;
	XtSetArg(args[n], XtNwidth, 140); n++;
	Won = XtCreateManagedWidget("On", commandWidgetClass, buttonbox, args, n);
	XtAddCallback(Won, XtNcallback, On, NULL);
	Woff = XtCreateManagedWidget("Off", commandWidgetClass, buttonbox, args, n);
	XtAddCallback(Woff, XtNcallback, Off, NULL);
	Wupdate = XtCreateManagedWidget("Update", commandWidgetClass, buttonbox, args, n);
	XtAddCallback(Wupdate, XtNcallback, Update, NULL);
	Wread = XtCreateManagedWidget("ReadCurrent", commandWidgetClass, buttonbox, args, n);
	XtAddCallback(Wread, XtNcallback, ReadHit, NULL);
	Wadd = XtCreateManagedWidget("AddRow", commandWidgetClass, buttonbox, args, n);
	XtAddCallback(Wadd, XtNcallback, AddRow, NULL);
	Wquit = XtCreateManagedWidget("Quit", commandWidgetClass, buttonbox, args, n);
	XtAddCallback(Wquit, XtNcallback, Quit, NULL);

	XtAppAddActions(app_context, Window_actions, XtNumber(Window_actions));

	/* SynchroScroll */
	{
		SynchroWho *tvlwho, *vlwho, *tvrwho, *vrwho;

		tvlwho = MakeSynchroWho(titleviewleft);
		vlwho = MakeSynchroWho(viewleft);
		tvrwho = MakeSynchroWho(titleviewright);
		vrwho = MakeSynchroWho(viewright);

		SynchroPair(tvlwho, vlwho, FALSE);
		SynchroPair(tvrwho, vrwho, FALSE);
		SynchroPair(tvlwho, tvrwho, TRUE);
		SynchroPair(vlwho, vrwho, TRUE);
	}

	XtRealizeWidget(topLevel);

	/* Clean up */
	CamouflageBar(titleviewleft, TRUE);
	CamouflageBar(titleviewleft, FALSE);
	CamouflageBar(viewleft, TRUE);
	CamouflageBar(titleviewright, FALSE);

	XawViewportSetLocation(titleviewright, -1.0, .5);

	/* Start the update timer */
	XtAppAddTimeOut(app_context, 1000/PER_SECOND,
		(XtTimerCallbackProc)TimeoutHandler, NULL);

	XtAppMainLoop(app_context);
}

static void
On(Widget w, XtPointer client_data, XtPointer call_data)
{
	if ( hiton()) {
		perror("hitbox on");
	}
}

static void
Off(Widget w, XtPointer client_data, XtPointer call_data)
{
	if ( hitoff()) {
		perror("hitbox off");
	}
}


static void
Update(Widget w, XtPointer client_data, XtPointer call_data)
{
	int i, ret;

	for (i=1; i < table_size; ++i) {
		if (hitstats[i].changed) {
			/*printf("Updating entry %d\n", i);*/
			ret = addhit(row1_values[0][i].value.string, /* src */
				row1_values[1][i].value.string, /* dest */
				row1_values[2][i].value.number, /* delay */
				row1_values[3][i].value.number, /* delsigma */
				row1_values[4][i].value.number, /* bandwidth */
				row1_values[5][i].value.number, /* drop */
				row1_values[6][i].value.number, /* dup */
				row1_values[7][i].value.number, /* drdmin */
				row1_values[8][i].value.number, /* drdmax */
				&hitstats[i].stats.hitreq
			);
			if (ret < 0) {
				fprintf(stderr, "Error adding %s -> %s: ",
					row1_values[0][i].value.string,
					row1_values[1][i].value.string);
				if (h_errno) herror("addhit");
				else perror("addhit");
			}
			fixname(row1_values[0][i].value.string,
				hitstats[i].stats.hitreq.src);
			MyXawTextSetString(row1_values[0][i].widget,
				row1_values[0][i].value.string);
			fixname(row1_values[1][i].value.string,
				hitstats[i].stats.hitreq.dest);
			MyXawTextSetString(row1_values[1][i].widget,
				row1_values[1][i].value.string);
			hitstats[i].changed = FALSE;
			if (ret >= 0)
				hitstats[i].inuse = TRUE;
		}
	}

	RedrawAll(TRUE);
	return;
}


static void
Quit(Widget w, XtPointer client_data, XtPointer call_data)
{

	XtCloseDisplay(theDisplay);
	exit(0);
}


static void QuitAction(Widget w, XEvent *event, String *params, Cardinal *num)
{
	XtCloseDisplay(theDisplay);
	exit(0);
}


char *malcopy(char *string)
{
	char *answer = (char *)malloc(strlen(string) + 20);

	strcpy(answer, string);
	return answer;
}

static void
get_text(Widget w, XtPointer client_data, XtPointer call_data)
{
	char *str;
	int value, oldvalue;
	struct value_widget *entry;

	entry = (struct value_widget *)client_data;
	XtVaGetValues(w, XtNstring, &str, NULL);
	switch (entry->type) {
	case IntegerValue:
	case UsecValue:
	case ShortpercentValue:
		value = atovalue(str, entry->type);
		oldvalue = entry->value.number;
		if (oldvalue != value) {
			entry->changeme->changed = TRUE;
			/*printf("Value change %d to %d\n", oldvalue, value);*/
			entry->value.number = value;
			MyXawTextSetNumber(entry->widget, value, entry->type, FALSE);
		}
		break;
	case StringValue:
		if (strcmp(entry->value.string, str)) {
			entry->changeme->changed = TRUE;
			/*printf("String changed %s to %s\n",
				entry->value.string, str);*/
			free(entry->value.string);
			entry->value.string = malcopy(str);
		}
		break;
	}
}

static void
MyXawTextSetString(Widget textwidget, char *string)
{
	int n, oldlen, newlen;
	Arg args[4];
	XawTextPosition insert_point;
	char *setme;

	oldlen = strlen(string);
	if (oldlen < 5) oldlen = 5;
	oldlen *= 2;
	setme = (char *)malloc(oldlen);
	strcpy(setme, string);
	n = 0;
	XtSetArg(args[n], XtNstring, setme); n++;
	XtSetValues(textwidget, args, n);
	XawTextInvalidate(textwidget, (XawTextPosition) 0,
		(XawTextPosition) strlen(setme));
}

static void
MyXawTextSetNumber(Widget textwidget, unsigned int value, Value type, Boolean justify)
{
	int n, oldlen, newlen;
	Arg args[4];
	char text[20];
	char *oldstr;
	XawTextPosition insert_point;

	XtVaGetValues(textwidget, XtNstring, &oldstr, NULL);
	oldlen = strlen(oldstr);
	insert_point = XawTextGetInsertionPoint(textwidget);
	if (justify) {
		/* Kids, don't try this at home -- adjusting insertion point
		 * relative to end of string!
		 */

		switch (type) {
		case IntegerValue:
			sprintf(text, "%11u", value);
			break;
		case UsecValue:
			sprintf(text,"%7u.%03u", value/THOUSAND, value%THOUSAND);
			break;
		case ShortpercentValue:
			sprintf(text, "%11.4f", ((double)value*100.0)/PROBFACTOR);
			break;
		}
		newlen = strlen(text);
		if (newlen != oldlen) {
			insert_point += newlen - oldlen;
			if (insert_point > newlen)
				insert_point = newlen;
			/*printf("\tLength change %d to %d, insert at %d\n",
				oldlen, strlen(text), insert_point);*/
		}

	} else {
		strcpy(text, oldstr);
	}

	n = 0;
	XtSetArg(args[n], XtNstring, text); n++;
	XtSetValues(textwidget, args, n);
	XawTextInvalidate(textwidget, (XawTextPosition) 0,
		(XawTextPosition) strlen(text));
	XawTextSetInsertionPoint(textwidget, insert_point);
}

/* Utility functions */

/* Take time in decimal msecs, convert to usecs */

int
atousec(char *string)
{
	double	mseconds;

	mseconds = atof(string);

	return (int)(mseconds*(float)THOUSAND);
}

/* Take probability in percent form, convert to scale integer */

int
atoshortpercent(char *string)
{
	double percent;

	percent = atof(string);
	return (int) ((percent*(float)PROBFACTOR)/100.0);
}

int
atovalue(char *string, Value type)
{
	switch (type) {
	case IntegerValue:
		return atoi(string);
	case UsecValue:
		return atousec(string);	
	case ShortpercentValue:
		return atoshortpercent(string);	
	}
}
